# coding=utf-8

"""Define hooks for code_aster repositories."""

import os.path as osp
import re

from ..ext_utils import get_changed_files, get_repo_name, is_admin
from .generic import (NOOK, OK, branch_checker, branch_user_commit_hook,
                      branch_user_push_hook, check_commit_msg)

# authorized branches names (+ vNN)
AUTH_BRANCH = ('^default$', '^edf/', '^asterxx/',
               '^v([0-9]+(\.[0-9]+)?(_smeca)?)$')


# pylint: disable=invalid-name
repo_branch_checker = branch_checker(authorized=AUTH_BRANCH)

pretxncommit = branch_user_commit_hook(repo_branch_checker, is_admin,
                                       check_msg=True)

pretxnchangegroup = branch_user_push_hook(repo_branch_checker, is_admin,
                                          check_msg=True)


def aslint(repo, paths=()):
    """Call aslint before a commit: in case of known error/warning, try to
    fix it. The commit command is not cancelled.
    """
    from subprocess import call
    from aslint.logger import logger
    from aslint.config import ASCFG
    from aslint.filetype import filter_dict
    from aslint.base_checkers import Report, checkcontext
    from aslint.check_files import check_files, read_waf_parameters
    # global informations
    checkcontext.reponame = get_repo_name(repo.root)
    checkcontext.branch = repo['.'].branch()

    # logger.title("running aslint...")
    choice = ASCFG.get('check.precommit')
    if choice == "no":
        logger.warn("pre-commit checkings skipped because of 'check.precommit'"
                    " value in your ~/.hgrc file")
        return OK
    elif choice == "undef":
        logger.info("pre-commit: If you don't plan to submit your work "
                    "you can disable these checkings by adding "
                    "'check.precommit = no' in the [aster] section in "
                    "your ~/.hgrc file.")
        logger.info("pre-commit: To hide this message you can set the value "
                    "'check.precommit = yes'.")
    elif choice != "yes":
        logger.warn("unknown choice {0!r} for 'check.precommit'"
                    " (see ~/.hgrc) must be 'yes' or 'no'.".format(choice))
    dchg = get_changed_files(repo.root, files=paths)
    changes = filter_dict(dchg)
    report = Report()
    wafc4che = osp.join(ASCFG.get('waf.builddir'), 'c4che', 'release_cache.py')
    c4che = read_waf_parameters(wafc4che, prefer_sequential=True)

    check_files(report, changes, c4che)
    if report.hasError():
        # run the fixers and remove the messages from the report
        dcmd = report.getFixerCommands()
        for eid, cmd in list(dcmd.items()):
            logger.warn("fixing error {0}...".format(eid))
            call(cmd, shell=True)
            report.remove(eid)
    # check for remaining errors
    if not report.hasError():
        logger.info("pre-commit: aslint hook passed successfully.")
    else:
        logger.info(report.to_text())
        report.report_summary()
        answ = input("The commit will be not accepted because of remaining "
                     "errors.\nDo you want to continue (y/n) ?")
        if answ.lower() not in ('y', 'o'):
            logger.info("pre-commit: aslint ended with errors.")
            return NOOK
        else:
            logger.cancel_error()
    return OK
